from __future__ import annotations

__all__ = ("OperationsQueue",)

import math
import os
import queue
from collections.abc import Hashable
from multiprocessing import Queue
from time import monotonic
from typing import (
    TYPE_CHECKING,
    Optional,
)

from neptune_scale.api.validation import verify_type
from neptune_scale.exceptions import NeptuneUnableToLogData
from neptune_scale.sync.parameters import (
    MAX_MULTIPROCESSING_QUEUE_SIZE,
    MAX_QUEUE_ELEMENT_SIZE,
    MAX_QUEUE_SIZE,
)
from neptune_scale.sync.queue_element import SingleOperation
from neptune_scale.util import (
    envs,
    get_logger,
)
from neptune_scale.util.abstract import Resource

if TYPE_CHECKING:
    from threading import RLock

    from neptune_api.proto.neptune_pb.ingest.v1.pub.ingest_pb2 import RunOperation

logger = get_logger()

# We use this value when comparing time since the last successful put. This is needed to
# avoid cases where the actual wait time is lower than the `timeout` due to the resolution of the monotonic clock.
# The call to put(block=True, timeout=...) could end fractions of a second earlier than the requested timeout. This
# happens eg. on Windows.
# The next call would then block again unnecessarily, if made quickly enough, even though the previous one failed:
# t=0.000s: successful put to empty queue, last_put_time -> 0.000s
# t=0.010s: failed put with timeout=2s, which blocks and ends early after 1.980s
# t=1.995s: monotonic() - last_put_time == 1.995s, which is still less than 2s -> we would block again,
# but we shouldn't
MONOTONIC_CLOCK_RESOLUTION_UPPER_BOUND = 0.1


class OperationsQueue(Resource):
    def __init__(
        self,
        *,
        lock: RLock,
        max_size: int = MAX_QUEUE_SIZE,
    ) -> None:
        verify_type("max_size", max_size, int)

        self._lock: RLock = lock
        self._max_size: int = max_size

        self._sequence_id: int = 0
        self._last_timestamp: Optional[float] = None
        self._queue: Queue[SingleOperation] = Queue(maxsize=min(MAX_MULTIPROCESSING_QUEUE_SIZE, max_size))
        self._last_successful_put_time = monotonic()

        self._max_blocking_time = envs.get_int(envs.LOG_MAX_BLOCKING_TIME_SECONDS, None) or math.inf
        if self._max_blocking_time < 0:
            raise ValueError(f"{envs.LOG_MAX_BLOCKING_TIME_SECONDS} must be a non-negative number.")

        action = os.getenv(envs.LOG_FAILURE_ACTION, "drop")
        if action not in ("drop", "raise"):
            raise ValueError(f"Invalid value '{action}' for {envs.LOG_FAILURE_ACTION}. Must be 'drop' or 'raise'.")

        self._raise_on_enqueue_failure = action == "raise"

    @property
    def queue(self) -> Queue[SingleOperation]:
        return self._queue

    @property
    def last_sequence_id(self) -> int:
        with self._lock:
            return self._sequence_id - 1

    @property
    def last_timestamp(self) -> Optional[float]:
        with self._lock:
            return self._last_timestamp

    def enqueue(self, *, operation: RunOperation, size: Optional[int] = None, key: Hashable = None) -> None:
        try:
            is_metadata_update = operation.HasField("update")
            serialized_operation = operation.SerializeToString()

            if len(serialized_operation) > MAX_QUEUE_ELEMENT_SIZE:
                raise ValueError(f"Operation size exceeds the maximum allowed size ({MAX_QUEUE_ELEMENT_SIZE})")

            with self._lock:
                self._last_timestamp = monotonic()

                item = SingleOperation(
                    sequence_id=self._sequence_id,
                    timestamp=self._last_timestamp,
                    operation=serialized_operation,
                    metadata_size=size,
                    is_batchable=is_metadata_update,
                    batch_key=key,
                )

                # Optimistically put the item without blocking. If the queue is full, we will retry
                # the put with blocking, but only if the last successful put was lest than the `timeout` ago.
                # This way if the sync process is stuck, we will drop operations until we are able to successfully
                # put an item into the queue again after some of the pending items were processed.
                try:
                    self._queue.put_nowait(item)
                    self._last_successful_put_time = monotonic()
                except queue.Full:
                    elapsed_since_last_put_success = (
                        monotonic() - self._last_successful_put_time + MONOTONIC_CLOCK_RESOLUTION_UPPER_BOUND
                    )
                    if elapsed_since_last_put_success < self._max_blocking_time:
                        try:
                            self._queue.put(item, block=True, timeout=self._max_blocking_time)
                            self._last_successful_put_time = monotonic()
                        except queue.Full:
                            self._on_enqueue_failed("Operations queue is full", operation)
                            return
                    else:
                        self._on_enqueue_failed("Operations queue is full", operation)
                        return

                self._sequence_id += 1
        # Pass this through, as it's raised intentionally in _on_enqueue_failed()
        except NeptuneUnableToLogData as e:
            raise e
        except Exception as e:
            self._on_enqueue_failed(reason=str(e), operation=operation)

    def close(self) -> None:
        self._queue.close()
        # This is needed to avoid hanging the main process
        self._queue.cancel_join_thread()

    def _on_enqueue_failed(self, reason: str, operation: RunOperation) -> None:
        if self._raise_on_enqueue_failure:
            raise NeptuneUnableToLogData(reason=reason, operation=str(operation))
        else:
            logger.error(f"Dropping operation due to error: {reason}. Operation: {operation}")



from __future__ import annotations

__all__ = ("MetadataSplitter",)

import math
import warnings
from collections.abc import (
    Callable,
    Iterator,
)
from datetime import datetime
from typing import (
    Any,
    Optional,
    TypeVar,
    Union,
)

from more_itertools import peekable
from neptune_api.proto.neptune_pb.ingest.v1.common_pb2 import (
    SET_OPERATION,
    Preview,
    UpdateRunSnapshot,
    Value,
)
from neptune_api.proto.neptune_pb.ingest.v1.pub.ingest_pb2 import RunOperation

from neptune_scale.api.metrics import Metrics
from neptune_scale.exceptions import (
    NeptuneFloatValueNanInfUnsupported,
    NeptuneScaleWarning,
)
from neptune_scale.net.serialization import (
    datetime_to_proto,
    make_step,
    make_value,
    pb_key_size,
)
from neptune_scale.util import (
    envs,
    get_logger,
)

logger = get_logger()

T = TypeVar("T", bound=Any)


class MetadataSplitter(Iterator[tuple[RunOperation, int]]):
    def __init__(
        self,
        *,
        project: str,
        run_id: str,
        timestamp: datetime,
        configs: Optional[dict[str, Union[float, bool, int, str, datetime, list, set, tuple]]],
        metrics: Optional[Metrics],
        add_tags: Optional[dict[str, Union[list[str], set[str], tuple[str]]]],
        remove_tags: Optional[dict[str, Union[list[str], set[str], tuple[str]]]],
        max_message_bytes_size: int = 1024 * 1024,
    ):
        self._should_skip_non_finite_metrics = envs.get_bool(envs.SKIP_NON_FINITE_METRICS, True)
        self._step = make_step(number=metrics.step) if (metrics is not None and metrics.step is not None) else None
        self._timestamp = datetime_to_proto(timestamp)
        self._project = project
        self._run_id = run_id
        self._configs = peekable(configs.items()) if configs else None
        self._metrics_data = (
            peekable(self._skip_non_finite(metrics.step, metrics.data)) if metrics is not None else None
        )
        self._metrics_preview = metrics.preview if metrics is not None else False
        self._metrics_preview_completion = metrics.preview_completion if metrics is not None else 0.0
        self._add_tags = peekable(add_tags.items()) if add_tags else None
        self._remove_tags = peekable(remove_tags.items()) if remove_tags else None

        self._max_update_bytes_size = (
            max_message_bytes_size
            - RunOperation(
                project=self._project,
                run_id=self._run_id,
                update=self._make_empty_update_snapshot(),
            ).ByteSize()
        )
        self._has_returned = False

    def __iter__(self) -> MetadataSplitter:
        self._has_returned = False
        return self

    def __next__(self) -> tuple[RunOperation, int]:
        update = self._make_empty_update_snapshot()
        size = update.ByteSize()

        size = self.populate(
            assets=self._configs,
            update_producer=lambda key, value: update.assign[key].MergeFrom(value),
            size=size,
        )
        size = self.populate(
            assets=self._metrics_data,
            update_producer=lambda key, value: update.append[key].MergeFrom(value),
            size=size,
        )
        size = self.populate_tags(
            update=update,
            assets=self._add_tags,
            operation=SET_OPERATION.ADD,
            size=size,
        )
        size = self.populate_tags(
            update=update,
            assets=self._remove_tags,
            operation=SET_OPERATION.REMOVE,
            size=size,
        )

        if not self._has_returned or update.assign or update.append or update.modify_sets:
            self._has_returned = True
            return RunOperation(project=self._project, run_id=self._run_id, update=update), size
        else:
            raise StopIteration

    def populate(
        self,
        assets: Optional[peekable[Any]],
        update_producer: Callable[[str, Value], None],
        size: int,
    ) -> int:
        if not assets:
            return size

        while size < self._max_update_bytes_size:
            try:
                key, value = assets.peek()
            except StopIteration:
                break

            proto_value = make_value(value)
            new_size = size + pb_key_size(key) + proto_value.ByteSize() + 6

            if new_size > self._max_update_bytes_size:
                break

            update_producer(key, proto_value)
            size, _ = new_size, next(assets)

        return size

    def populate_tags(
        self, update: UpdateRunSnapshot, assets: Optional[peekable[Any]], operation: SET_OPERATION.ValueType, size: int
    ) -> int:
        if not assets:
            return size

        while size < self._max_update_bytes_size:
            try:
                key, values = assets.peek()
            except StopIteration:
                break

            if not isinstance(values, peekable):
                values = peekable(values)

            is_full = False
            new_size = size + pb_key_size(key) + 6
            for value in values:
                tag_size = pb_key_size(value) + 6
                if new_size + tag_size > self._max_update_bytes_size:
                    values.prepend(value)
                    is_full = True
                    break

                update.modify_sets[key].string.values[value] = operation
                new_size += tag_size

            size, _ = new_size, next(assets)

            if is_full:
                assets.prepend((key, list(values)))
                break

        return size

    def _skip_non_finite(self, step: Optional[float | int], metrics: dict[str, float]) -> Iterator[tuple[str, float]]:
        """Yields (metric, value) pairs, skipping non-finite values depending on the env setting."""

        for k, v in metrics.items():
            v = float(v)

            if not math.isfinite(v):
                if self._should_skip_non_finite_metrics:
                    warnings.warn(
                        f"Neptune is skipping non-finite metric values. You can turn this warning into an error by "
                        f"setting the `{envs.SKIP_NON_FINITE_METRICS}` environment variable to `False`.",
                        category=NeptuneScaleWarning,
                        stacklevel=7,
                    )

                    logger.warning(f"Skipping a non-finite value `{v}` of metric `{k}` at step `{step}`. ")
                    continue
                else:
                    raise NeptuneFloatValueNanInfUnsupported(metric=k, step=step, value=v)

            yield k, v

    def _make_empty_update_snapshot(self) -> UpdateRunSnapshot:
        include_preview = self._metrics_data and self._metrics_preview
        return UpdateRunSnapshot(
            step=self._step,
            preview=(self._make_preview() if include_preview else None),
            timestamp=self._timestamp,
            assign={},
            append={},
            modify_sets={},
        )

    def _make_preview(self) -> Optional[Preview]:
        if not self._metrics_preview:
            return None
        # let backend default completion
        if self._metrics_preview_completion is not None:
            return Preview(is_preview=True, completion_ratio=self._metrics_preview_completion)
        return Preview(is_preview=True)
